@@ -1,6 +1,6 @@
 //
 // MIT License
-// Copyright (c) 2019 Jonathan R. Madsen
+// Copyright (c) 2020 Jonathan R. Madsen
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
 // in the Software without restriction, including without limitation the rights
@@ -29,133 +29,196 @@
 // ---------------------------------------------------------------
 
 #include "PTL/ThreadPool.hh"
-#include "PTL/Globals.hh"
+#include "PTL/GetEnv.hh"
+#include "PTL/ScopeDestructor.hh"
 #include "PTL/ThreadData.hh"
+#include "PTL/Threading.hh"
 #include "PTL/UserTaskQueue.hh"
 #include "PTL/VUserTaskQueue.hh"
 
-#include <cstdlib>
-
-#if defined(PTL_USE_GPERF)
-#    include <gperftools/heap-checker.h>
-#    include <gperftools/heap-profiler.h>
-#    include <gperftools/profiler.h>
-#endif
-
-//======================================================================================//
-
-inline intmax_t
-ncores()
-{
-    return static_cast<intmax_t>(Thread::hardware_concurrency());
-}
-
-//======================================================================================//
-
-ThreadPool::thread_id_map_t ThreadPool::f_thread_ids;
+#include <cassert>
+#include <mutex>
+#include <new>
+#include <stdexcept>
+#include <thread>
 
 //======================================================================================//
 
 namespace
 {
-ThreadData::pointer_t&
+PTL::ThreadData*&
 thread_data()
 {
-    return ThreadData::GetInstance();
+    return PTL::ThreadData::GetInstance();
 }
+}  // namespace
+
+namespace PTL
+{
+//======================================================================================//
+
+ThreadPool::thread_id_map_t&
+ThreadPool::f_thread_ids()
+{
+    static auto _v = thread_id_map_t{};
+    return _v;
 }
 
 //======================================================================================//
 
-bool ThreadPool::f_use_tbb = false;
+ThreadPool::size_type&
+ThreadPool::f_default_pool_size()
+{
+    static size_type _v =
+        GetEnv<size_type>("PTL_NUM_THREADS", Thread::hardware_concurrency());
+    return _v;
+}
 
 //======================================================================================//
 // static member function that calls the member function we want the thread to
 // run
 void
-ThreadPool::start_thread(ThreadPool* tp, intmax_t _idx)
+ThreadPool::start_thread(ThreadPool* tp, thread_data_t* _data, intmax_t _idx)
 {
+    if(tp->get_verbose() > 0)
+    {
+        AutoLock lock(TypeMutex<decltype(std::cerr)>());
+        std::cerr << "[PTL::ThreadPool] Starting thread " << _idx << "..." << std::endl;
+    }
+
+    auto _thr_data = std::make_shared<ThreadData>(tp);
     {
         AutoLock lock(TypeMutex<ThreadPool>(), std::defer_lock);
         if(!lock.owns_lock())
             lock.lock();
         if(_idx < 0)
-            _idx = f_thread_ids.size();
-        f_thread_ids[std::this_thread::get_id()] = _idx;
+            _idx = f_thread_ids().size();
+        f_thread_ids()[std::this_thread::get_id()] = _idx;
+        SetThreadId(_idx);
+        _data->emplace_back(_thr_data);
     }
-    thread_data().reset(new ThreadData(tp));
+    thread_data() = _thr_data.get();
+    tp->record_entry();
     tp->execute_thread(thread_data()->current_queue);
+    tp->record_exit();
+
+    if(tp->get_verbose() > 0)
+    {
+        AutoLock lock(TypeMutex<decltype(std::cerr)>());
+        std::cerr << "[PTL::ThreadPool] Thread " << _idx << " terminating..."
+                  << std::endl;
+    }
 }
 
 //======================================================================================//
-// static member function that initialized tbb library
-void
-ThreadPool::set_use_tbb(bool enable)
+
+const ThreadPool::thread_id_map_t&
+ThreadPool::get_thread_ids()
 {
-#if defined(PTL_USE_TBB)
-    f_use_tbb = enable;
-#else
-    ConsumeParameters<bool>(enable);
-#endif
+    return f_thread_ids();
 }
 
 //======================================================================================//
 
 uintmax_t
-ThreadPool::GetThisThreadID()
+ThreadPool::get_thread_id(ThreadId _tid)
 {
-    auto _tid = ThisThread::get_id();
+    uintmax_t _idx = 0;
     {
         AutoLock lock(TypeMutex<ThreadPool>(), std::defer_lock);
         if(!lock.owns_lock())
             lock.lock();
-        if(f_thread_ids.find(_tid) == f_thread_ids.end())
+        auto itr = f_thread_ids().find(_tid);
+        if(itr == f_thread_ids().end())
         {
-            auto _idx          = f_thread_ids.size();
-            f_thread_ids[_tid] = _idx;
+            _idx                 = f_thread_ids().size();
+            f_thread_ids()[_tid] = _idx;
+        }
+        else
+        {
+            _idx = itr->second;
         }
     }
-    return f_thread_ids[_tid];
+    return _idx;
 }
 
 //======================================================================================//
 
-ThreadPool::ThreadPool(const size_type& pool_size, VUserTaskQueue* task_queue,
-                       bool _use_affinity, const affinity_func_t& _affinity_func)
-: m_use_affinity(_use_affinity)
-, m_tbb_tp(false)
-, m_alive_flag(false)
-, m_verbose(0)
-, m_pool_size(0)
-, m_pool_state(thread_pool::state::NONINIT)
-, m_master_tid(ThisThread::get_id())
-, m_thread_awake(new atomic_int_type(0))
-, m_task_queue(task_queue)
-, m_tbb_task_group(nullptr)
-, m_init_func([]() { return; })
-, m_affinity_func(_affinity_func)
+uintmax_t
+ThreadPool::get_this_thread_id()
 {
-    m_verbose = GetEnv<int>("PTL_VERBOSE", m_verbose);
+    return get_thread_id(ThisThread::get_id());
+}
 
-    if(!m_task_queue)
-        m_task_queue = new UserTaskQueue(pool_size);
+//======================================================================================//
 
-    auto master_id = GetThisThreadID();
+uintmax_t
+ThreadPool::add_thread_id(ThreadId _tid)
+{
+    AutoLock lock(TypeMutex<ThreadPool>(), std::defer_lock);
+    if(!lock.owns_lock())
+        lock.lock();
+    if(f_thread_ids().find(_tid) == f_thread_ids().end())
+    {
+        auto _idx            = f_thread_ids().size();
+        f_thread_ids()[_tid] = _idx;
+        SetThreadId(_idx);
+    }
+    return f_thread_ids().at(_tid);
+}
+
+//======================================================================================//
+
+ThreadPool::ThreadPool(const Config& _cfg)
+: m_use_affinity{ _cfg.use_affinity }
+, m_tbb_tp{ _cfg.use_tbb }
+, m_verbose{ _cfg.verbose }
+, m_priority{ _cfg.priority }
+, m_pool_state{ std::make_shared<std::atomic_short>(thread_pool::state::NONINIT) }
+, m_task_queue{ _cfg.task_queue }
+, m_init_func{ _cfg.initializer }
+, m_fini_func{ _cfg.finalizer }
+, m_affinity_func{ _cfg.set_affinity }
+{
+    auto master_id = get_this_thread_id();
     if(master_id != 0 && m_verbose > 1)
-        std::cerr << "ThreadPool created on non-master slave" << std::endl;
+    {
+        AutoLock lock(TypeMutex<decltype(std::cerr)>());
+        std::cerr << "[PTL::ThreadPool] ThreadPool created on worker thread" << std::endl;
+    }
 
-    thread_data().reset(new ThreadData(this));
+    thread_data() = new ThreadData(this);
 
-    // initialize after GetThisThreadID so master is zero
-    this->initialize_threadpool(pool_size);
+    // initialize after get_this_thread_id so master is zero
+    if(_cfg.init)
+        this->initialize_threadpool(_cfg.pool_size);
 }
 
 //======================================================================================//
 
 ThreadPool::~ThreadPool()
 {
-    if(m_alive_flag.load())
-        destroy_threadpool();
+    if(m_alive_flag->load())
+    {
+        std::cerr << "Warning! ThreadPool was not properly destroyed! Call "
+                     "destroy_threadpool() before deleting the ThreadPool object to "
+                     "eliminate this message."
+                  << std::endl;
+        m_pool_state->store(thread_pool::state::STOPPED);
+        m_task_lock->lock();
+        m_task_cond->notify_all();
+        m_task_lock->unlock();
+        for(auto& itr : m_threads)
+            itr.join();
+        m_threads.clear();
+    }
+
+    // delete owned resources
+    if(m_delete_task_queue)
+        delete m_task_queue;
+
+    delete m_tbb_task_arena;
+    delete m_tbb_task_group;
 }
 
 //======================================================================================//
@@ -163,13 +226,29 @@ ThreadPool::~ThreadPool()
 bool
 ThreadPool::is_initialized() const
 {
-    return !(m_pool_state.load() == thread_pool::state::NONINIT);
+    return !(m_pool_state->load() == thread_pool::state::NONINIT);
 }
 
 //======================================================================================//
 
 void
-ThreadPool::set_affinity(intmax_t i, Thread& _thread)
+ThreadPool::record_entry()
+{
+    ++(*m_thread_active);
+}
+
+//======================================================================================//
+
+void
+ThreadPool::record_exit()
+{
+    --(*m_thread_active);
+}
+
+//======================================================================================//
+
+void
+ThreadPool::set_affinity(intmax_t i, Thread& _thread) const
 {
     try
     {
@@ -177,15 +256,39 @@ ThreadPool::set_affinity(intmax_t i, Thread& _thread)
         intmax_t     _pin          = m_affinity_func(i);
         if(m_verbose > 0)
         {
-            std::cout << "Setting pin affinity for thread " << _thread.get_id() << " to "
-                      << _pin << std::endl;
+            AutoLock lock(TypeMutex<decltype(std::cerr)>());
+            std::cerr << "[PTL::ThreadPool] Setting pin affinity for thread "
+                      << get_thread_id(_thread.get_id()) << " to " << _pin << std::endl;
         }
-        Threading::SetPinAffinity(_pin, native_thread);
-    }
-    catch(std::runtime_error& e)
+        SetPinAffinity(_pin, native_thread);
+    } catch(std::runtime_error& e)
     {
-        std::cout << "Error setting pin affinity" << std::endl;
-        std::cerr << e.what() << std::endl;  // issue assigning affinity
+        std::cerr << "[PTL::ThreadPool] Error setting pin affinity: " << e.what()
+                  << std::endl;
+    }
+}
+
+//======================================================================================//
+
+void
+ThreadPool::set_priority(int _prio, Thread& _thread) const
+{
+    try
+    {
+        NativeThread native_thread = _thread.native_handle();
+        if(m_verbose > 0)
+        {
+            AutoLock lock(TypeMutex<decltype(std::cerr)>());
+            std::cerr << "[PTL::ThreadPool] Setting thread "
+                      << get_thread_id(_thread.get_id()) << " priority to " << _prio
+                      << std::endl;
+        }
+        SetThreadPriority(_prio, native_thread);
+    } catch(std::runtime_error& e)
+    {
+        AutoLock lock(TypeMutex<decltype(std::cerr)>());
+        std::cerr << "[PTL::ThreadPool] Error setting thread priority: " << e.what()
+                  << std::endl;
     }
 }
 
@@ -201,72 +304,52 @@ ThreadPool::initialize_threadpool(size_type proposed_size)
 
     //--------------------------------------------------------------------//
     // store that has been started
-    if(!m_alive_flag.load())
-        m_pool_state.store(thread_pool::state::STARTED);
+    if(!m_alive_flag->load())
+        m_pool_state->store(thread_pool::state::STARTED);
 
-        //--------------------------------------------------------------------//
-        // handle tbb task scheduler
-#ifdef PTL_USE_TBB
-    if(f_use_tbb)
+#if defined(PTL_USE_TBB)
+    //--------------------------------------------------------------------//
+    // handle tbb task scheduler
+    if(m_tbb_tp)
     {
         m_tbb_tp                               = true;
         m_pool_size                            = proposed_size;
-        tbb_task_scheduler_t*& _task_scheduler = tbb_task_scheduler();
+        tbb_global_control_t*& _global_control = tbb_global_control();
         // delete if wrong size
         if(m_pool_size != proposed_size)
         {
-            delete _task_scheduler;
-            _task_scheduler = nullptr;
+            delete _global_control;
+            _global_control = nullptr;
         }
 
-        if(!_task_scheduler)
+        if(!_global_control)
         {
-            _task_scheduler =
-                new tbb_task_scheduler_t(tbb::task_scheduler_init::deferred);
-        }
-
-        if(!_task_scheduler->is_active())
-        {
-            m_pool_size = proposed_size;
-            _task_scheduler->initialize(proposed_size + 1);
+            _global_control = new tbb_global_control_t(
+                tbb::global_control::max_allowed_parallelism, proposed_size + 1);
             if(m_verbose > 0)
             {
-                std::cout << "ThreadPool [TBB] initialized with " << m_pool_size
-                          << " threads." << std::endl;
+                AutoLock lock(TypeMutex<decltype(std::cerr)>());
+                std::cerr << "[PTL::ThreadPool] ThreadPool [TBB] initialized with "
+                          << m_pool_size << " threads." << std::endl;
             }
         }
+
         // create task group (used for async)
         if(!m_tbb_task_group)
-            m_tbb_task_group = new tbb_task_group_t();
-        return m_pool_size;
-    }
+        {
+            m_tbb_task_group = new tbb_task_group_t{};
+            execute_on_all_threads([this]() { m_init_func(); });
+        }
 
-    // NOLINT(readability-else-after-return)
-    if(f_use_tbb && tbb_task_scheduler())
-    {
-        m_tbb_tp                               = false;
-        tbb_task_scheduler_t*& _task_scheduler = tbb_task_scheduler();
-        if(_task_scheduler)
-        {
-            _task_scheduler->terminate();
-            delete _task_scheduler;
-            _task_scheduler = nullptr;
-        }
-        // delete task group (used for async)
-        if(m_tbb_task_group)
-        {
-            m_tbb_task_group->wait();
-            delete m_tbb_task_group;
-            m_tbb_task_group = nullptr;
-        }
+        return m_pool_size;
     }
 #endif
 
-    m_alive_flag.store(true);
+    m_alive_flag->store(true);
 
     //--------------------------------------------------------------------//
     // if started, stop some thread if smaller or return if equal
-    if(m_pool_state.load() == thread_pool::state::STARTED)
+    if(m_pool_state->load() == thread_pool::state::STARTED)
     {
         if(m_pool_size > proposed_size)
         {
@@ -274,8 +357,18 @@ ThreadPool::initialize_threadpool(size_type proposed_size)
                 ;
             if(m_verbose > 0)
             {
-                std::cout << "ThreadPool initialized with " << m_pool_size << " threads."
-                          << std::endl;
+                AutoLock lock(TypeMutex<decltype(std::cerr)>());
+                std::cerr << "[PTL::ThreadPool] ThreadPool initialized with "
+                          << m_pool_size << " threads." << std::endl;
+            }
+            if(!m_task_queue)
+            {
+                m_delete_task_queue = true;
+                m_task_queue        = new UserTaskQueue(m_pool_size);
+            }
+            else
+            {
+                m_task_queue->resize(m_pool_size);
             }
             return m_pool_size;
         }
@@ -283,9 +376,15 @@ ThreadPool::initialize_threadpool(size_type proposed_size)
         {
             if(m_verbose > 0)
             {
-                std::cout << "ThreadPool initialized with " << m_pool_size << " threads."
+                AutoLock lock(TypeMutex<decltype(std::cerr)>());
+                std::cerr << "ThreadPool initialized with " << m_pool_size << " threads."
                           << std::endl;
             }
+            if(!m_task_queue)
+            {
+                m_delete_task_queue = true;
+                m_task_queue        = new UserTaskQueue(m_pool_size);
+            }
             return m_pool_size;
         }
     }
@@ -293,45 +392,53 @@ ThreadPool::initialize_threadpool(size_type proposed_size)
     //--------------------------------------------------------------------//
     // reserve enough space to prevent realloc later
     {
-        AutoLock _task_lock(m_task_lock);
+        AutoLock _task_lock(*m_task_lock);
         m_is_joined.reserve(proposed_size);
     }
 
-    auto this_tid = GetThisThreadID();
+    if(!m_task_queue)
+    {
+        m_delete_task_queue = true;
+        m_task_queue        = new UserTaskQueue(proposed_size);
+    }
+
+    auto this_tid = get_this_thread_id();
     for(size_type i = m_pool_size; i < proposed_size; ++i)
     {
         // add the threads
         try
         {
-            using pointer_t = std::unique_ptr<Thread>;
-            auto tid =
-                pointer_t(new Thread(ThreadPool::start_thread, this, this_tid + i + 1));
+            // create thread
+            Thread thr{ ThreadPool::start_thread, this, &m_thread_data,
+                        this_tid + i + 1 };
             // only reaches here if successful creation of thread
             ++m_pool_size;
             // store thread
-            m_main_threads.push_back(tid->get_id());
+            m_main_threads.push_back(thr.get_id());
             // list of joined thread booleans
             m_is_joined.push_back(false);
             // set the affinity
             if(m_use_affinity)
-                set_affinity(i, *tid.get());
-            // detach
-            m_unique_threads.emplace_back(std::move(tid));
-        }
-        catch(std::runtime_error& e)
+                set_affinity(i, thr);
+            set_priority(m_priority, thr);
+            // store
+            m_threads.emplace_back(std::move(thr));
+        } catch(std::runtime_error& e)
         {
-            std::cerr << e.what() << std::endl;  // issue creating thread
+            AutoLock lock(TypeMutex<decltype(std::cerr)>());
+            std::cerr << "[PTL::ThreadPool] " << e.what()
+                      << std::endl;  // issue creating thread
             continue;
-        }
-        catch(std::bad_alloc& e)
+        } catch(std::bad_alloc& e)
         {
-            std::cerr << e.what() << std::endl;
+            AutoLock lock(TypeMutex<decltype(std::cerr)>());
+            std::cerr << "[PTL::ThreadPool] " << e.what() << std::endl;
             continue;
         }
     }
     //------------------------------------------------------------------------//
 
-    AutoLock _task_lock(m_task_lock);
+    AutoLock _task_lock(*m_task_lock);
 
     // thread pool size doesn't match with join vector
     // this will screw up joining later
@@ -347,9 +454,11 @@ ThreadPool::initialize_threadpool(size_type proposed_size)
 
     if(m_verbose > 0)
     {
-        std::cout << "ThreadPool initialized with " << m_pool_size << " threads."
-                  << std::endl;
+        AutoLock lock(TypeMutex<decltype(std::cerr)>());
+        std::cerr << "[PTL::ThreadPool] ThreadPool initialized with " << m_pool_size
+                  << " threads." << std::endl;
     }
+
     return m_main_threads.size();
 }
 
@@ -358,43 +467,56 @@ ThreadPool::initialize_threadpool(size_type proposed_size)
 ThreadPool::size_type
 ThreadPool::destroy_threadpool()
 {
-    int nid = static_cast<int>(GetThisThreadID());
-
     // Note: this is not for synchronization, its for thread communication!
     // destroy_threadpool() will only be called from the main thread, yet
     // the modified m_pool_state may not show up to other threads until its
     // modified in a lock!
     //------------------------------------------------------------------------//
-    m_pool_state.store(thread_pool::state::STOPPED);
-
-    //------------------------------------------------------------------------//
-    // notify all threads we are shutting down
-    m_task_lock.lock();
-    m_task_cond.notify_all();
-    m_task_lock.unlock();
+    m_pool_state->store(thread_pool::state::STOPPED);
 
     //--------------------------------------------------------------------//
     // handle tbb task scheduler
-#ifdef PTL_USE_TBB
-    if(m_tbb_tp && tbb_task_scheduler())
-    {
-        tbb_task_scheduler_t*& _task_scheduler = tbb_task_scheduler();
-        delete _task_scheduler;
-        _task_scheduler = nullptr;
-        m_tbb_tp        = false;
-        std::cout << "ThreadPool [TBB] destroyed" << std::endl;
-    }
+#if defined(PTL_USE_TBB)
     if(m_tbb_task_group)
     {
-        m_tbb_task_group->wait();
+        execute_on_all_threads([this]() { m_fini_func(); });
+        auto _func = [&]() { m_tbb_task_group->wait(); };
+        if(m_tbb_task_arena)
+            m_tbb_task_arena->execute(_func);
+        else
+            _func();
         delete m_tbb_task_group;
         m_tbb_task_group = nullptr;
     }
+    if(m_tbb_task_arena)
+    {
+        delete m_tbb_task_arena;
+        m_tbb_task_arena = nullptr;
+    }
+    if(m_tbb_tp && tbb_global_control())
+    {
+        tbb_global_control_t*& _global_control = tbb_global_control();
+        delete _global_control;
+        _global_control = nullptr;
+        m_tbb_tp        = false;
+        AutoLock lock(TypeMutex<decltype(std::cerr)>());
+        if(m_verbose > 0)
+        {
+            std::cerr << "[PTL::ThreadPool] ThreadPool [TBB] destroyed" << std::endl;
+        }
+    }
 #endif
 
-    if(!m_alive_flag.load())
+    if(!m_alive_flag->load())
         return 0;
 
+    //------------------------------------------------------------------------//
+    // notify all threads we are shutting down
+    m_task_lock->lock();
+    m_task_cond->notify_all();
+    m_task_lock->unlock();
+    //------------------------------------------------------------------------//
+
     if(m_is_joined.size() != m_main_threads.size())
     {
         std::stringstream ss;
@@ -405,28 +527,76 @@ ThreadPool::destroy_threadpool()
         throw std::runtime_error(ss.str());
     }
 
-    //--------------------------------------------------------------------//
-    // erase thread from thread ID list
-    for(auto _tid : m_main_threads)
+    for(size_type i = 0; i < m_is_joined.size(); i++)
     {
-        if(f_thread_ids.find(_tid) != f_thread_ids.end())
-            f_thread_ids.erase(f_thread_ids.find(_tid));
+        //--------------------------------------------------------------------//
+        //
+        if(i < m_threads.size())
+            m_threads.at(i).join();
+
+        //--------------------------------------------------------------------//
+        // if its joined already, nothing else needs to be done
+        if(m_is_joined.at(i))
+            continue;
+
+        //--------------------------------------------------------------------//
+        // join
+        if(std::this_thread::get_id() == m_main_threads[i])
+            continue;
+
+        //--------------------------------------------------------------------//
+        // thread id and index
+        auto _tid = m_main_threads[i];
+
+        //--------------------------------------------------------------------//
+        // erase thread from thread ID list
+        if(f_thread_ids().find(_tid) != f_thread_ids().end())
+            f_thread_ids().erase(f_thread_ids().find(_tid));
+
+        //--------------------------------------------------------------------//
+        // it's joined
+        m_is_joined.at(i) = true;
     }
 
-    //--------------------------------------------------------------------//
-    // try waking up a bunch of threads that are still waiting
-    m_task_cond.notify_all();
-
-    for(auto& itr : m_unique_threads)
-        itr->join();
-
+    m_thread_data.clear();
+    m_threads.clear();
     m_main_threads.clear();
     m_is_joined.clear();
-    m_unique_threads.clear();
 
-    m_alive_flag.store(false);
+    m_alive_flag->store(false);
 
-    printf("[%i]> ThreadPool destroyed...\n", nid);
+    auto start   = std::chrono::steady_clock::now();
+    auto elapsed = std::chrono::duration<double>{};
+    // wait maximum of 30 seconds for threads to exit
+    while(m_thread_active->load() > 0 && elapsed.count() < 30)
+    {
+        std::this_thread::sleep_for(std::chrono::milliseconds(50));
+        elapsed = std::chrono::steady_clock::now() - start;
+    }
+
+    auto _active = m_thread_active->load();
+
+    if(get_verbose() > 0)
+    {
+        if(_active == 0)
+        {
+            AutoLock lock(TypeMutex<decltype(std::cerr)>());
+            std::cerr << "[PTL::ThreadPool] ThreadPool destroyed" << std::endl;
+        }
+        else
+        {
+            AutoLock lock(TypeMutex<decltype(std::cerr)>());
+            std::cerr << "[PTL::ThreadPool] ThreadPool destroyed but " << _active
+                      << " threads might still be active (and cause a termination error)"
+                      << std::endl;
+        }
+    }
+
+    if(m_delete_task_queue)
+    {
+        delete m_task_queue;
+        m_task_queue = nullptr;
+    }
 
     return 0;
 }
@@ -436,19 +606,24 @@ ThreadPool::destroy_threadpool()
 ThreadPool::size_type
 ThreadPool::stop_thread()
 {
-    if(!m_alive_flag.load() || m_pool_size == 0)
+    if(!m_alive_flag->load() || m_pool_size == 0)
         return 0;
 
+    m_pool_state->store(thread_pool::state::PARTIAL);
+
     //------------------------------------------------------------------------//
     // notify all threads we are shutting down
-    m_task_lock.lock();
+    m_task_lock->lock();
     m_is_stopped.push_back(true);
-    m_task_cond.notify_one();
-    m_task_lock.unlock();
+    m_task_cond->notify_one();
+    m_task_lock->unlock();
     //------------------------------------------------------------------------//
 
+    while(!m_is_stopped.empty() && m_stop_threads.empty())
+        ;
+
     // lock up the task queue
-    AutoLock _task_lock(m_task_lock);
+    AutoLock _task_lock(*m_task_lock);
 
     while(!m_stop_threads.empty())
     {
@@ -468,44 +643,65 @@ ThreadPool::stop_thread()
         m_is_joined.pop_back();
     }
 
+    m_pool_state->store(thread_pool::state::STARTED);
+
     m_pool_size = m_main_threads.size();
     return m_main_threads.size();
 }
 
 //======================================================================================//
 
+ThreadPool::task_queue_t*&
+ThreadPool::get_valid_queue(task_queue_t*& _queue) const
+{
+    if(!_queue)
+        _queue = new UserTaskQueue{ static_cast<intmax_t>(m_pool_size) };
+    return _queue;
+}
+//======================================================================================//
+
 void
 ThreadPool::execute_thread(VUserTaskQueue* _task_queue)
 {
-#if defined(PTL_USE_GPERF)
-    ProfilerRegisterThread();
-#endif
-
-    // how long the thread waits on condition variable
-    // static int wait_time = GetEnv<int>("PTL_POOL_WAIT_TIME", 5);
-
     ++(*m_thread_awake);
 
     // initialization function
     m_init_func();
+    // finalization function (executed when scope is destroyed)
+    ScopeDestructor _fini{ [this]() { m_fini_func(); } };
 
-    ThreadId tid  = ThisThread::get_id();
-    auto&    data = thread_data();
+    ThreadId    tid  = ThisThread::get_id();
+    ThreadData* data = thread_data();
     // auto        thread_bin = _task_queue->GetThreadBin();
     // auto        workers    = _task_queue->workers();
 
+    auto start   = std::chrono::steady_clock::now();
+    auto elapsed = std::chrono::duration<double>{};
+    // check for updates for 60 seconds max
+    while(!_task_queue && elapsed.count() < 60)
+    {
+        elapsed = std::chrono::steady_clock::now() - start;
+        data->update();
+        _task_queue = data->current_queue;
+    }
+
+    if(!_task_queue)
+    {
+        --(*m_thread_awake);
+        throw std::runtime_error("No task queue was found after 60 seconds!");
+    }
+
     assert(data->current_queue != nullptr);
     assert(_task_queue == data->current_queue);
 
     // essentially a dummy run
+    if(_task_queue)
     {
         data->within_task = true;
         auto _task        = _task_queue->GetTask();
         if(_task)
         {
             (*_task)();
-            if(!_task->group())
-                delete _task;
         }
         data->within_task = false;
     }
@@ -513,19 +709,23 @@ ThreadPool::execute_thread(VUserTaskQueue* _task_queue)
     // threads stay in this loop forever until thread-pool destroyed
     while(true)
     {
+        static thread_local auto p_task_lock = m_task_lock;
+
         //--------------------------------------------------------------------//
         // Try to pick a task
-        AutoLock _task_lock(m_task_lock, std::defer_lock);
+        AutoLock _task_lock(*p_task_lock, std::defer_lock);
         //--------------------------------------------------------------------//
 
         auto leave_pool = [&]() {
-            auto _state      = [&]() { return static_cast<int>(m_pool_state.load()); };
+            auto _state      = [&]() { return static_cast<int>(m_pool_state->load()); };
             auto _pool_state = _state();
             if(_pool_state > 0)
             {
                 // stop whole pool
                 if(_pool_state == thread_pool::state::STOPPED)
                 {
+                    if(_task_lock.owns_lock())
+                        _task_lock.unlock();
                     return true;
                 }
                 // single thread stoppage
@@ -537,6 +737,9 @@ ThreadPool::execute_thread(VUserTaskQueue* _task_queue)
                     {
                         m_stop_threads.push_back(tid);
                         m_is_stopped.pop_back();
+                        if(_task_lock.owns_lock())
+                            _task_lock.unlock();
+                        // exit entire function
                         return true;
                     }
                     if(_task_lock.owns_lock())
@@ -554,7 +757,7 @@ ThreadPool::execute_thread(VUserTaskQueue* _task_queue)
         //    actually true!
         while(_task_queue->empty())
         {
-            auto _state = [&]() { return static_cast<int>(m_pool_state.load()); };
+            auto _state = [&]() { return static_cast<int>(m_pool_state->load()); };
             auto _size  = [&]() { return _task_queue->true_size(); };
             auto _empty = [&]() { return _task_queue->empty(); };
             auto _wake  = [&]() { return (!_empty() || _size() > 0 || _state() > 0); };
@@ -564,7 +767,7 @@ ThreadPool::execute_thread(VUserTaskQueue* _task_queue)
 
             if(_task_queue->true_size() == 0)
             {
-                if(m_thread_awake && m_thread_awake->load() > 0)
+                if(m_thread_awake->load() > 0)
                     --(*m_thread_awake);
 
                 // lock before sleeping on condition
@@ -574,10 +777,9 @@ ThreadPool::execute_thread(VUserTaskQueue* _task_queue)
                 // Wait until there is a task in the queue
                 // Unlocks mutex while waiting, then locks it back when signaled
                 // use lambda to control waking
-                m_task_cond.wait(_task_lock, _wake);
+                m_task_cond->wait(_task_lock, _wake);
 
-                // leave the pool immediately
-                if(leave_pool())
+                if(_state() == thread_pool::state::STOPPED)
                     return;
 
                 // unlock if owned
@@ -585,7 +787,7 @@ ThreadPool::execute_thread(VUserTaskQueue* _task_queue)
                     _task_lock.unlock();
 
                 // notify that is awake
-                if(m_thread_awake && m_thread_awake->load() < m_pool_size)
+                if(m_thread_awake->load() < m_pool_size)
                     ++(*m_thread_awake);
             }
             else
@@ -595,12 +797,12 @@ ThreadPool::execute_thread(VUserTaskQueue* _task_queue)
         // release the lock
         if(_task_lock.owns_lock())
             _task_lock.unlock();
+
         //----------------------------------------------------------------//
 
         // leave pool if conditions dictate it
         if(leave_pool())
             return;
-        //----------------------------------------------------------------//
 
         // activate guard against recursive deadlock
         data->within_task = true;
@@ -613,8 +815,6 @@ ThreadPool::execute_thread(VUserTaskQueue* _task_queue)
             if(_task)
             {
                 (*_task)();
-                if(!_task->group())
-                    delete _task;
             }
         }
         //----------------------------------------------------------------//
@@ -626,3 +826,5 @@ ThreadPool::execute_thread(VUserTaskQueue* _task_queue)
 }
 
 //======================================================================================//
+
+}  // namespace PTL
